/*--------------------------------------------------
Pinbelegung:

Vin                 Rot1 (Vcc Ultraschallsensor)
Gnd (egal wo)      Schwarz1 (Gnd Ultraschallsensor)

Vin                 Rot2 (Vcc LEDs)
Gnd (egal wo)       Schwarz2 (Gnd LEDs)

DIP 1 (s.o.)         Rot3 (Servo links)
Gnd (egal wo)       Schwarz3 (Servo links)

DIP 2 (s.o.)         Rot4 (Servo rechts)
Gnd (egal wo)       Schwarz4 (Servo rechts)

D0                  Unten am Board FEST mit Schalter 3 und 4 verbunden
D1                  Orange kurz (SCL LEDs)
D2                  Blau (SDA LEDs)
D5                  Grau lang (Trigger Ultraschallsensor)
D6                  Blau lang (Echo Ultraschallsensor)
D7                  Weiß (Servo rechts)
D8                  Grün (Servo links)

Vorsicht: D0-GPIO16, D1-GPIO5, D2-GPIO4, D3-GPIO0, D4-GPIO2, D5-GPIO14, D6-GPIO12, D7-GPIO13, D8-GPIO15, D9-GPIO3, D10-GPIO1, LED:GPIO12, SDA: GPIO4->D2, SCL:GPIO5->D1
Auf NodeMCUs-Pins (Dx) kann direkt mit hier genannten Namen zugegriffen werden (siehe "pins_arduino.h")
D3 und D4 werden außerdem benutzt beim Flashen. Können belegt werden, müssen aber zum Flashen herausgezogen werden.

DIP-Schalter:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! VOR DER FAHRT
- Beide Stromleitungen der Servos auf Eingänge von 1 legen
-> Durch Schalten von 1 wird die Versorgung der Servos auf Vin geschaltet
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! SONST KEIN LOSFAHREN MÖGLICH!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! VOR DER FAHRT AUSWÄHLEN
-> Durch Schalten von 3 oder 4 wird zwischen Steuerung mittels Beschleunigungssensor oder mittels Maus/Touchevent gewählt
--> Steht 3 oder 4 auf on, wird D0 auf low gezogen --> Steuerung mittels dem Beschleunigungssensor
--> Stehen beide auf 0 ist D0 high (über internen Pullup, siehe unten) --> Steuerung mit Touch-/Mausevent
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! VOR DER FAHRT AUSWÄHLEN

1: Eingang gebrückt mit 2, Ausgang ist unten verbunden mit Vin
2: keine Funktion
3: Eingang ist unten verbunden mit D0, Ausgang ist verbunden mit Gnd
4: gebrückt, wie 3

Webseite:
192.168.4.1 in Browser öffnen (Vorzugsweise Google Chrome)

--------------------------------------------------*/

// Filesystem Init
#include "FS.h"

// WiFi Init
#include <ESP8266WiFi.h>
const char* ssid = "Legoino";
const char* password = "";
String WebsiteBody, WebsiteHeader;
WiFiServer server(80);

// LED-Array Init
#include <Wire.h>
#include "Adafruit_LEDBackpack.h"
#include "Adafruit_GFX.h"
Adafruit_8x8matrix matrix = Adafruit_8x8matrix();

// Servo Init
#include <Servo.h> 
const int servoZeroR = 90, servoZeroL = 90;
const int servoDelta = 30; 
int servoSpeedL = servoZeroR, servoSpeedR = servoZeroR;
int direcFlagAct = 0;
Servo myservoL, myservoR;

// Initialwerte für die Steuerung
double x_val=0, y_val=0;
//    Richtungen
//       1 2 3    vorne
//       4 5 6    mitte
//       7 8 9    hinten
// links-mitte-rechts
int Richtungswahl=5;

// Ultraschallsensor
int Trigger = D5;
int EchoPin = D6;
unsigned long echo = 0;
int sensorWert = 0;
unsigned long letzterZeitwert = 0;
int letzterSensorWert=0;
// Warte auf Echo. 5000 entspricht 86 cm Höchstlänge. Falls mehr gewünscht, reagiert Steuerung entsprechend langsamer
int waitingTime=5000;
int ZeitAusrufezeichen=0;

// Allgemeine Inits
// Eigens definierte Bitmaps
#include <Bitmaps.h>
// Min-Operator
#define min(a,b) ((a)<(b)?(a):(b))
// Debug: Wenn true: Ausgaben Serial
bool debug=false;
// Webseite seriell anzeigen
bool WebseiteAnzeigen=false;
// GET-Request anzeigen
bool ClientRequestShow=false;
// Pins können "D0...Dn" gennant werden und werden automatisch in GPIO-Pins umgewandelt
#include <pins_arduino.h>

void setup() 
{
  // Ultraschall-Pins
  pinMode(Trigger,OUTPUT);
  pinMode(EchoPin,INPUT);
  
  // Seriellen Kanal öffnen
  Serial.begin(9600);
  delay(1);
  
  // Srver in AP Modus starten
  WiFi.mode(WIFI_AP);
  WiFi.softAP(ssid, password);
  server.begin();

  // HTML-Seite aus Speicher holen
  SPIFFS.begin();
  File f= SPIFFS.open("/header.txt","r");
  WebsiteHeader = f.readString();
  f.close();
  pinMode(D0, INPUT_PULLUP);
  
  // D0 ist defaultmäßig high. Kann auf Betätigung von Taster 3 oder 4 low gezogen werden.
  if (digitalRead(D0)==LOW){
    f= SPIFFS.open("/accel.txt","r");
    WebsiteBody = f.readString();
    f.close();
    if (debug) Serial.println("\naccel");
  }
  // Ansonsten: Maus/TouchEvent-Version laden
  else{
    f= SPIFFS.open("/mouse.txt","r");
    WebsiteBody = f.readString();
    f.close();
    if (debug) Serial.println("\nmouse");
  }
  

  // Webseite anzeigen
  if(debug&&WebseiteAnzeigen){
    Serial.println(WebsiteBody);
  }

  // Attach Servos, auf 0 setzen
  myservoR.attach(D7);  
  myservoL.attach(D8);
  myservoL.write(servoZeroL);
  myservoR.write(servoZeroR);
  delay(100);

  // LED-Matrix starten
  matrix.begin(0x70);
  matrix.setRotation(3);

  // Okay-Bitmap zeigen
  matrix.clear();
  matrix.drawBitmap(0, 0, okay_bmp, 8, 8, LED_ON);
  matrix.writeDisplay();

}

void loop() 
{ 
  // Ist ein Client vorhanden?
  WiFiClient client = server.available();
  if (!client) 
  {
    return;
  }
  
  // Client bekommt Zeitfenster um Daten zu senden
 if (debug) Serial.println("new client");
 unsigned long ultimeout = millis()+250;
 while(!client.available() && (millis()<ultimeout) )
 {
   delay(1);
 }
 if(millis()>ultimeout) 
 { 
   if (debug) Serial.println("client connection time-out!");
   return; 
 }

  // Read the first line of the request
  String ClientRequest = client.readStringUntil('\r');
  if (debug) Serial.println(ClientRequest);
  client.flush();

  // Wandle String in Char-Array um darin mit strstr suchen zu können
  int laenge=ClientRequest.length();
  char ClientRequestChar[laenge];
  strcpy(ClientRequestChar, ClientRequest.c_str());



  // GET-Request enthält <<KEINEN>> String "param"
  if(!strstr(ClientRequestChar,"param")){
    if (debug) Serial.println("No parameters");
    //client.print(WebsiteHeader+WebsiteBody.length()+"\n\n"+WebsiteBody);
    client.print(WebsiteHeader+WebsiteBody.length()+"\n\n"+WebsiteBody);
    if (debug && WebseiteAnzeigen){
      Serial.print(WebsiteHeader+WebsiteBody.length()+"\n\n"+WebsiteBody);
    }
  }
  // GET-Request ENTHÄLT String "param"
  else{
    // Nachricht kommt in der Form:
    // GET /?param=x:X.XXXX,y:Y.YYYY; HTTP/1.1 -> XPOS und YPOS sind die im Browser angezeigten Zahlen
    // x und y aus ClientRequest herausfischen
    Serial.println("Parameter gefunden");
    String x_Str=ClientRequest.substring(ClientRequest.indexOf("x:")+2,ClientRequest.indexOf(","));
    double x_val=x_Str.toFloat();
    String y_Str=ClientRequest.substring(ClientRequest.indexOf("y:")+2,ClientRequest.indexOf(";"));
    double y_val=y_Str.toFloat();

    // Anzeigen lassen zum Debuggen
    if (debug && ClientRequestShow){
      Serial.println("x: " + ClientRequest.substring(ClientRequest.indexOf("x:")+2,ClientRequest.indexOf(",")));
      Serial.println("y: " + ClientRequest.substring(ClientRequest.indexOf("y:")+2,ClientRequest.indexOf(";")));
    }
  
    // Move je nach X und Y
    x_val*=90;
    y_val*=90;
    
    //    Richtungen
    //       1 2 3    vorne
    //       4 5 6    mitte
    //       7 8 9    hinten
    // links-mitte-rechts
    if (y_val<-20){
      if (x_val<-20) Richtungswahl=1;
      else if (x_val>=-20 && x_val <=20) Richtungswahl=2;
      else if (x_val>20) Richtungswahl=3;
    }
    else if (y_val>=-20 && y_val <=20){
      if (x_val<-20) Richtungswahl=4;
      else if (x_val>=-20 && x_val <=20) Richtungswahl=5;
      else if (x_val>20) Richtungswahl=6;
    }
    else if (y_val>20){
      if (x_val<-20) Richtungswahl=7;
      else if (x_val>=-20 && x_val <=20) Richtungswahl=8;
      else if (x_val>20) Richtungswahl=9;
    }

    // Zeichne Richtungwahl auf LED-Array
    matrix.clear();
    switch (Richtungswahl){
      case 1:
          matrix.drawBitmap(0, 0, vor_links_bmp, 8, 8, LED_ON);
          break;
      case 2:
          matrix.drawBitmap(0, 0, vor_bmp, 8, 8, LED_ON);
          break;
      case 3:
          matrix.drawBitmap(0, 0, vor_rechts_bmp, 8, 8, LED_ON);
        break;
      case 4:
          matrix.drawBitmap(0, 0, links_bmp, 8, 8, LED_ON);
        break;
      case 5:
          matrix.drawBitmap(0, 0, stehen_bmp, 8, 8, LED_ON);
        break;
      case 6:
          matrix.drawBitmap(0, 0, rechts_bmp, 8, 8, LED_ON);
        break;
      case 7:
          matrix.drawBitmap(0, 0, zurueck_links_bmp, 8, 8, LED_ON);
          break;
      case 8:
          matrix.drawBitmap(0, 0, zurueck_bmp, 8, 8, LED_ON);
          break;
      case 9:
          matrix.drawBitmap(0, 0, zurueck_rechts_bmp, 8, 8, LED_ON);
        break;
    }

    // Fahrzeug sehr nah an Wand: Ausrufezeichen malen und stehen bleiben
    // Alle 500 ms wird zwischen "leer" und "Ausrufezeichen" gewechselt
    // Vorsicht: Sensorwert 0: ungültig
    if (sensorWert>0&&sensorWert<=20&&(Richtungswahl<=3)){
      Richtungswahl=5;
      matrix.clear();
      matrix.drawBitmap(0, 0, ausruf_bmp, 8, 8, LED_ON);
      if (millis()-ZeitAusrufezeichen>=500){
        ZeitAusrufezeichen=millis();
        matrix.clear();
        matrix.drawBitmap(0, 0, all_off_bmp, 8, 8, LED_ON);
      }
    }
    matrix.writeDisplay();
    moove(Richtungswahl);

    // Sensor nur auslesen, wenn 300ms seit der letzten Messung
    // Warum? Sensorwert wird nur alle 400ms vom Server angefragt -> Öftere Messung nicht erforderlich
    if((millis()-letzterZeitwert>=300))
    {
      letzterZeitwert=millis();
      letzterSensorWert=sensorWert;
      sensorWert = ping();
      if (debug) Serial.println(sensorWert);
    }
    // Falls Sensorwert=0 -> Wert ungültig, Sensorwert nicht updaten
    if (sensorWert!=0) client.print(sensorWert);
    // Filterung, falls kein Sensorwert ankommt
    else if (letzterSensorWert>0){
      client.print(letzterSensorWert);
      sensorWert=letzterSensorWert;
    }
  }
  
  // Client schließen
  client.stop();
  if (debug) Serial.println("Client disonnected");
}

// Fahrfunktion
void moove(int direcFlag) {
  //Nur eingreifen, wenn sich die Sollfahrtrichtung geändert hat.
  if (direcFlag != direcFlagAct){
    switch(direcFlag){
    case 1:
      //vorne links
      while((servoSpeedR != (servoZeroR+servoDelta))||(servoSpeedL != servoZeroL)){
        servoSpeedR = servoSpeedR + 3*sign((servoZeroR+servoDelta)-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign(servoZeroL-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 1;
      break;
    case 2:
      //vorwärts
      while((servoSpeedR != (servoZeroR+servoDelta))||(servoSpeedL != (servoZeroL-servoDelta))){
        servoSpeedR = servoSpeedR + 3*sign((servoZeroR+servoDelta)-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign((servoZeroL-servoDelta)-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 2;
      break;
    case 3:
      //vorne rechts
      while((servoSpeedR != servoZeroR)||(servoSpeedL != (servoZeroL-servoDelta))){
        servoSpeedR = servoSpeedR + 3*sign(servoZeroR-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign((servoZeroL-servoDelta)-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 3;
      break;
    case 4:
      //voll links
      while((servoSpeedR != (servoZeroR+servoDelta))||(servoSpeedL != (servoZeroL+servoDelta))){
        servoSpeedR = servoSpeedR + 3*sign((servoZeroR+servoDelta)-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign((servoZeroL+servoDelta)-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 4;
      break;
    case 5:
      //stopp
      while((servoSpeedR != servoZeroR)||(servoSpeedL != servoZeroL)){
        servoSpeedR = servoSpeedR + 3*sign(servoZeroR-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign(servoZeroL-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 5;
      break;
    case 6:
      //voll rechts
      while((servoSpeedR != (servoZeroR-servoDelta))||(servoSpeedL != (servoZeroL-servoDelta))){
        servoSpeedR = servoSpeedR + 3*sign((servoZeroR-servoDelta)-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign((servoZeroL-servoDelta)-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 6;
      break;
    case 7:
      //zurück links
      while((servoSpeedR != servoZeroR)||(servoSpeedL != (servoZeroL+servoDelta))){
        servoSpeedR = servoSpeedR + 3*sign(servoZeroR-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign((servoZeroL+servoDelta)-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 7;
      break;
    case 8:
      //voll zurück
      while((servoSpeedR != (servoZeroR-servoDelta))||(servoSpeedL != (servoZeroL+servoDelta))){
        servoSpeedR = servoSpeedR + 3*sign((servoZeroR-servoDelta)-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign((servoZeroL+servoDelta)-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 8;
      break;
    case 9:
      //zurück rechts
      while((servoSpeedR != (servoZeroR-servoDelta))||(servoSpeedL != servoZeroL)){
        servoSpeedR = servoSpeedR + 3*sign((servoZeroR-servoDelta)-servoSpeedR);
        servoSpeedL = servoSpeedL + 3*sign(servoZeroL-servoSpeedL);

        myservoR.write(servoSpeedR);
        myservoL.write(servoSpeedL);
        delay(50);
      }
      direcFlagAct = 9;
      break;
    }
  }  
}

// Signum-Funktion
int sign (int a){
  if (a ==0) return 0;
  else return (a/abs(a));
}

// Ping-Funktion
unsigned long ping()
{ 
  digitalWrite(Trigger, LOW);                   // Send low pulse 
  delayMicroseconds(2);                         // Wait for 2 microseconds
  digitalWrite(Trigger, HIGH);                  // Send high pulse
  delayMicroseconds(5);                         // Wait for 5 microseconds
  digitalWrite(Trigger, LOW);                   // Holdoff
  digitalWrite(Trigger, HIGH);                  // Turn on pullup resistor
  echo = pulseIn(EchoPin, HIGH,waitingTime);    //Listen for echo
  sensorWert = (echo / 58.138);                 //convert to CM
  return sensorWert;
}

